; Z8671	TINY BASIC
; reconstructed	Volker Pohlers 10/2021,	Germany, Neustadt i.H.
; modified for TINY ES40

; THE IL CONTROL SECTION

X1	BR STRT		     ; ENTRY POINT DIRECT MODE
X2	BR GOTO		     ; ENTRY POINT NEXT	STMT MODE
X3	BR XEC		     ; ENTRY POINT RUN MODE
STRT	PC ":"		     ; PROMPT OUTPUT
	GL		     ; GET LINE
	BE X4		     ; BRANCH IF NOT EMPTY
	BR STRT		     ; TRY AGAIN IF NULL LINE
X4	BN GOTO		     ; TEST FOR	LINE NUMBER
	TS		     ; SKIP NEXT IF POSITIVE
	BR *		     ; ERROR STOP
		ILERR "Program line has a line number 0 or greater than 32768."
	SB		     ; SAVE BASIC POINTER
	DS		     ; DUPLICATE TOP OF	STACK
	GO		     ; GOTO LINE
	SN		     ; PUT FOUND LINE NUMBER ON	STACK
	SP		     ; STACK POP
	IL		     ; INSERT LINE INTO	PROGRAM
		ILERR "Memory full; new line not inserted."
	BR STRT		     ; GO GET NEXT
XEC	SB		     ; SAVE POINTERS FOR RUN WITH
	RB		     ; CONCATENATED INPUT
	LN 00		     ; PUT 0000	ON STACK
	GO		     ; GOTO FIRST LINE
	SN		     ; PUT FOUND LINE NUMBER ON	STACK
	TS		     ; SKIP NEXT IF POSITIVE
	BR *		     ; ERROR STOP
		ILERR "No program to RUN."
	SP		     ; STACK POP
X5	XQ		     ; EXECUTE PROGRAM

; STATEMENT EXECUTOR

GOTO	BC IF, "GO"	     ;
	BC GOSB,"TO"	     ; GOTO
S0	JS EXPR		     ; GET LINE	NUMBER
S1	BE *		     ; ERR IF NOT END OF LINE
		ILERR "GOTO is not at the end of program line."
	SB		     ; (DO THIS	FOR STARTING)
	RB		     ;
	TS		     ; SKIP NEXT IF POSITIVE
	BR *		     ; ERROR STOP
		ILERR "Cannot GOTO negative or zero line number."
	GO		     ; GO THERE
	BR X5		     ;
	BR *		     ;
		ILERR "Line number in GOTO does not exist."
IF	BC LET, "IF"	     ;
	JS REL		     ;
	BC S2, "THEN"	     ; OPTIONAL	NOISEWORD
S2	CP		     ; COMPARE SKIPS NEXT IF TRUE
	NX		     ; FALSE.
	BN S4		     ;
S3	BR S1		     ;
S4	BR GOTO		     ;
GOSB	BC S5, "SUB"	     ; GOSUB
	JS EXPR		     ; GET DESTINATION.
	BE *		     ; ERR IF NOT END OF LINE
		ILERR "GOSUB is not at the end of the line."
	SN		     ; PUT LINE	NUMBER ON STACK
	GS		     ;
	BR S3		     ;
S5	BC *, "@"	     ; GO@
		ILERR "Unrecognizable statement type beginning with GO."
	JS ARG3		     ; GET 3 ARGUMENTS
	BR S10		     ; EXECUTE MACHINE CODE
LET	BC S11, "LET"	     ;
LET1	BV S8		     ; VARIABLE	NAME?
S7	BC *, "="	     ;
		ILERR "Unrecognizable statement type, or '=' missing from LET statement."
	JS EXPR		     ; GO GET EXPRESSION
	SV		     ; STORE RESULT
	BR S11		     ;
S8	BC S9, "^"	     ; ADDR REFERENCE?
	JS FACT		     ; GET ADDR
	BR S7		     ;
S9	BC P0, "@"	     ; BYTE/REGISTER REFERENCE?
	LN LN81	     	     ; 0081 / =LO(LN81), HI(LN81) MUST BE 0
	JS FACT		     ; GET DEST
	BC *, "="	     ;
		ILERR "LET is missing its '='."
	JS EXPR		     ; GET SRC
S10	US		     ; USER SUBROUTINE.	STACK: 0081 (ADDR), DEST, SRC
	SP		     ; STACK POP
S11	BE S12		     ; IF STATEMENT END,
	NX		     ; DONE
S12	BC S14, ":"	     ; NEXT STATEMENT
S13	J  GOTO		     ; GET NEXT	STATEMENT
S14	BC PRNT, "\""	     ; PRINT STRING (WITHOUT KEYWORD)
	BR P8		     ;
PRNT	BC P9, "PRINT"	     ; PRINT
P0	BC P1, ":"	     ; STATEMENT END?
	NL		     ; NEW LINE
	BR S13		     ; NEXT STATEMENT
P1	BE P7		     ; END OF LINE?
	BR P5		     ; YES
P2	BC P4, ";"	     ; NO CR
P3	BE P6		     ; END OF LINE?
	NX		     ; YES, NEXT BASIC STATEMENT
P4	BC P5, ","	     ; COMMA SPACING?
	PT		     ; PRINT TAB
	BR P3		     ;
P5	NL		     ; NEW LINE
	BR S11		     ;
P6	BC P7, ":"	     ; PRINT STMT END?
	BR S13		     ;
P7	BC P10, "\""	     ; STRING?
P8	PQ		     ; PRINT QUOTE MARKS STRING
	BR *		     ; ERROR STOP
		ILERR "Quote missing in PRINT statement."
	BR P2		     ;
P9	BR RETN		     ; FORWARD BRANCH
P10	BC P12, "HEX("	     ; HEX?
	JS F12		     ; GET EXPR	AND )
	PH		     ; PRINT HEX NUMBER
P11	BR P2		     ;
P12	JS EXPR		     ; PRINT EXPR
	NE		     ; NEGATE
	TS		     ; SKIP NEXT IF POSITIVE
	BR P13		     ; NO, WAS POSITIVE	NUMBER
	PC "-"		     ; PRINT SIGN
	BR P14		     ;
P13	NE		     ; NEGATE (ORIG VALUE)
P14	PN		     ; PRINT NUMBER
	BR P11		     ;
RETN	BC STOP, "RET"	     ; RETURN
	BC S15, "URN"	     ; OPTIONAL
S15	BE *		     ; ERR IF NOT END OF LINE
		ILERR "RETURN is not at the end of the line."
	RS		     ; RESTORE SAVED LINE
		ILERR "GOSUB stack underflow."
	GO		     ;
	NX		     ; NEXT BASIC STATEMENT
	BR *		     ; ERROR STOP
		ILERR "The GOSUB for this RETURN no longer exists."
STOP	BC RUN, "STOP"	     ;
	BE S16		     ; END OF LINE?
		ILERR "STOP is not at the end of the line."
S16	WS		     ; YES, STOP
RUN	BC NEW, "RUN"	     ;
	J  XEC		     ; RUN PROGRAM
NEW	BC REM, "NEW"	     ;
	BE *		     ; ERR IF END OF LINE?
	MT		     ; MARK BASIC PGM EMPTY
REM	BC INPT, "REM"	     ;
	NX		     ; NEXT BASIC STATEMENT
INPT	BC LIST, "IN"	     ;
	BC S17, "PUT"	     ; INPUT
	BV *		     ; GET VARIABLE
		ILERR "INPUT variable name is missing."
	SB		     ; SWAP POINTERS
	BR I3		     ;
S17	BV *		     ; IN COMMAND, FIRST VAR?
		ILERR "IN or INPUT expects variable name."
	SB		     ; SWAP POINTERS
	BE I4		     ; END OF LINE?
I3	PC "? "		     ; YES, TYPE PROMPT
	GL		     ; READ INPUT LINE
	BE I4		     ; DID ANYTHING COME?
	BR I3		     ; NO, TRY AGAIN
I4	BC I5, ","	     ; NEXT PARAMETER (IN CMD)
I5	JS EXPR		     ; READ A NUMBER
	SV		     ; STORE INTO VARIABLE
	RB		     ; SWAP BACK
	BC I6, ","	     ; ANOTHER?
	BR S17		     ; YES IF COMMA
I6	J  S11		     ; OTHERWISE QUIT

LIST	BC L1, "LIST"	     ; LIST?
	SN		     ; PUT LINE	NUMBER ON STACK
	BE L2		     ; END OF LINE?
;	LN 0FFh		     ; THEN PUT	00FF ON	STACK
;	NE		     ; NEGATE -> FF01
;	LN 7Fh		     ; PUT 007F	ON STACK
;	BR L3		     ; FF01007F	-> 7F0100FF -> 7FFF0001
	LN 07FFFh
	LN 1
	BR L9
L1	J  LET1		     ; ELSE LET	STMT (WITHOUT KEYWORD)
L2	JS ARG2		     ; GET TWO NUMBERS
	BE *		     ; ERR IF NOT END OF LINE
		ILERR "LIST is not at end of line."
	SX 3		     ; EXCHANGE	THE TOP	TWO    aAbB > BAba
	SX 1		     ; NUMBERS ON THE STACK	    > BAab
L3	SX 3		     ;				    > bAaB
	SX 2		     ;				    > bBaA
L9	GO		     ; GO FIRST	LINE IN	AREA
	SN		     ; PUT LINE	NUMBER ON STACK
	SP		     ;
L4	NL		     ; OUTPUT NEW LINE
	CC		     ; GOT CANCEL CODE?
	BR L8		     ; THEN CANCEL LIST
	DS		     ; DUP STACK
	LB 03		     ; 03 (>=)
	SN		     ; PUT LINE	NUMBER ON STACK
	CP		     ; COMPARE.	SKIP NEXT IF TRUE
	BR L8		     ; ELSE LIST END
	SN		     ; PUT LINE	NUMBER ON STACK
	TS		     ; SKIP NEXT IF POSITIVE
	BR L7		     ; NEGATIVE	LINE NUMBER FFFF -> END
	PN		     ; OUTPUT LINE NO
	PC " "		     ; OUTPUT SPACE
L5	PQ		     ; OUTPUT LINE TO 00 OR "
	BR L6		     ; (PQ SKIPPES BRANCH IF ")
	PC "\""		     ; OUTPUT "
	BR L5		     ; AND CONTINUE
L6	FL		     ; SET NEXT LINE
	BR L4		     ; LOOP
L7	SP		     ; STACK POP
L8	SP		     ; RESTORE ORIG LINE NUMBER
	GO		     ; GO ORIG LINE NUMBER
	NX		     ;
	NX		     ; ?? WHY ??

; SUBROUTINES

REL	JS EXPR		     ; LEFT EXPRESSION
	BC R2, "<"	     ; CONVERT RELATION	OPERATORS
	BC R0, "="	     ; TO CODE BYTE ON STACK
	LB 06		     ; <=  6
	BR R5		     ;
R0	BC R1, ">"	     ;
	LB 05		     ; <>  5
	BR R5		     ;
R1	LB 04		     ; <   4
	BR R5		     ;
R2	BC R4, ">"	     ;
	BC R3, "="	     ;
	LB 03		     ; >=  3
	BR R5		     ;
R3	LB 01		     ; >   1
	BR R5		     ;
R4	BC *, "="	     ;
		ILERR "Unrecognizable relation in IF statement."
	LB 02		     ; =   2
R5	BR EXPR		     ; RIGHT EXPRESSION

ARG3	JS ARG2		     ; GET 3 ARGUMENTS
	BR A1		     ; 3RD OPT ARGUMENT
ARG2	JS EXPR		     ; GET 2 ARGUMENTS
A1	BC E6, ","	     ; OPTIONAL	ARGUMENT

EXPR	BC E0, "-"	     ; TRY FOR UNARY MINUS
	JS TERM		     ; GET VALUE
	NE		     ; NEGATE IT
	BR E1		     ; LOOK FOR	MORE
E0	BC E4, "+"	     ; IGNORE UNARY PLUS
E4	JS TERM		     ; GET VALUE
E1	BC E2, "+"	     ; TERMS SEPARATED BY PLUS
	JS TERM		     ; GET TERM
E5	AD		     ; SUM TERM
	BR E1		     ;
E2	BC E3, "-"	     ; TERMS SEPARATED BY MINUS
	JS TERM		     ; GET TERM
	NE		     ; NEGATE
	BR E5		     ; AND ADD
E6	DS		     ; DUPLICATE NUMBER	ON STACK
E3	RT		     ;

TERM	JS FACT		     ; LEFT FACTOR
		ILERR "Out of memory on GOSUB or expression evaluation."
T0	BC T1, "*"	     ; FACTORS SEPARATED BY TIMES
	JS FACT		     ; RIGHT FACTOR
	MP		     ; MULTIPLY
	BR T0		     ; NEXT FACTOR
T1	BC T6, "/"	     ; FACTORS SEPARATED BY DIVIDE
	NE		     ; NEGATE DIVIDEND
	TS		     ; SKIP NEXT IF POSITIVE
	BR T3		     ;
	JS FACT		     ; GET DIVISOR
	TS		     ; SKIP NEXT IF POSITIVE (-/+)
	BR T8		     ;
T2	DV		     ; DIVIDE (-/+, +/-)
	NE		     ; NEGATE RESULT
	BR T0		     ; NEXT FACTOR
T3	NE		     ; NEGATE DIVIDEND BACK
T4	JS FACT		     ; GET DIVISOR
	TS		     ; SKIP NEXT IF POSITIVE (+/+)
	BR T7		     ;
T5	DV		     ; DIVIDE  (+/+, -/-)
	BR T0		     ; NEXT FACTOR
T6	BC F3, "\\"	     ; UNSIGNED	DIVISION?
	BR T4		     ;
T7	NE		     ; NEGATE
	TS		     ; SKIP NEXT IF POSITIVE
	BR T9		     ;
	BR T2		     ; (+/-)
T8	NE		     ; NEGATE BACK
	TS		     ; SKIP NEXT IF POSITIVE
T9	BR *		     ; ERROR STOP
		ILERR "Divide by zero."
	BR T5		     ; (-/-)

FACT	BC F4, "USR("	     ; USR FUNCTION?
	JS ARG3		     ; GET ARGUMENTS
F1	BC *, ")"	     ;
		ILERR "Missing parenthesis in AND or USR call."
F2	US		     ; USER SUBROUTINE.	STACK USF-ADDR,	2 PARAM
F3	RT		     ; RETURN
F4	BC F5, "AND("	     ; AND FUNCTION?
	LN LN3C		     ; 003C ADDRESS USF	AND
F4A	JS ARG2		     ; GET ARGUMENTS
	BR F1		     ;
F5	BC F5A, "RND("
	LN SUB_RND
	BR F4A	
F5A	BC F6, "@"	     ; BYTE REFERENCE?
	LN LN47		     ; 0047 ADDRESS USF	FETCH
	JS FACT		     ; GET BYTE
	DS		     ; DUP STACK, DUMMY 2ND ARG
	BR F2		     ;
F6	BC F7, "^"	     ; ADDR REFERENCE?
	JS FACT		     ; GET ADDR
	BR F8		     ;
F7	BV F9		     ; VARIABLE?
F8	FV		     ; FETCH VALUE
	RT		     ; RETURN
F9	BC F10, "%"	     ; HEX NUMBER?
	HN		     ; GET HEX NUMBER
	RT		     ; RETURN
F10	BN F11		     ; NUMBER?
	RT		     ; GOT IT.
F11	BC *, "("	     ; OTHERWISE MUST BE (EXPR)
		ILERR "Syntax error in expression, or unrecognizable statement type."
F12	JS EXPR		     ;
	BC *, ")"	     ;
		ILERR "Missing right parenthesis in expression."
	RT		     ; RETURN
